/*
 * Copyright (c) 1998 by Groupe Bull. All Rights Reserved
 * Properties.java
 * $Id$
 */
package fr.dyade.koala.xml.util;

import org.xml.sax.DocumentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.Locator;
import org.xml.sax.AttributeList;
import org.xml.sax.Parser;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.ParserFactory;

import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * The <code>Properties</code> class represents a persistent set of 
 * properties. The <code>Properties</code> can be saved to a stream 
 * or loaded from a stream. Each key and its corresponding value in 
 * the property list is a string.</p>
 * <p>
 * A property list can contain another property list as its 
 * "defaults"; this second property list is searched if 
 * the property key is not found in the original property list.</p>
 * <p>
 * This is a special implementation for XML :</p>
 * <pre class="example">
 *   &lt;properties>
 *      &lt;key name="My_key1">My_Value1&lt;/key>
 *      &lt;key name="My_key2">My_Value2&lt;/key>
 *   &lt;/properties>
 * </pre>
 *
 * <ol>Current BUGS:
 *  <li>Handling for escaped characters doesn't work.
 * </ol>
 *
 * @author  Philippe Le Hgaret
 */
public class Properties extends java.util.Properties {
    
    /**
     * A property list that contains default values for any keys not 
     * found in this property list. 
     *
     * @since   JDK1.0
     */
    protected Properties defaults;
    
    /**
     * Creates an empty property list with no default values. 
     *
     * @since   JDK1.0
     */
    public Properties() {
	this(null);
    }
    
    /**
     * Creates an empty property list with the specified defaults. 
     *
     * @param   defaults   the defaults.
     * @since   JDK1.0
     */
    public Properties(Properties defaults) {
	this.defaults = defaults;
    }
    
    /**
     * Reads a property list from an input stream. 
     *
     * @param      in   the input stream.
     * @exception  IOException  if an error occurred when reading from the
     *               input stream.
     * @since   JDK1.0
     */
    public synchronized void load(InputStream in) throws IOException {
	PropertiesParser p = null;
	try {
	    p = new PropertiesParser(this, in);
	} catch (SAXException e) {
	    throw new IOException(e.toString());
	}
    }
    
    /**
     * Reads a property list from an input stream. 
     *
     * @param      in   the input stream.
     * @exception  IOException  if an error occurred when reading from the
     *               input stream.
     * @since   JDK1.0
     */
    synchronized void loadDefault(InputStream in) throws IOException {
	super.load(in);
    }
    
    /**
     * Stores this property list to the specified output stream. The 
     * string header is printed as a comment at the beginning of the stream.
     *
     * @param   out      an output stream.
     * @param   header   a description of the property list.
     * @since   JDK1.0
     */
    public synchronized void save(OutputStream out, String header) {
	PrintWriter wout = new PrintWriter(out);

	if (header != null) {
	    wout.print("<!--" + header + "-->");
	}
	
	Hashtable h = new Hashtable();
	enumerate(h);
	wout.print("<properties>");
	for (Enumeration e = h.keys() ; e.hasMoreElements() ;) {
	    String key = (String)e.nextElement();
	    String val = (String)h.get(key);
	    wout.print("  \n<key name=\"" + key + "\">");
	    if ((val.indexOf('&') != -1) || (val.indexOf('<') != -1)) {
		StringBuffer nval = new StringBuffer();
		for (int i = 0; i < val.length(); i++) {
		    char c = val.charAt(i);
		    if (c == '<') {
			nval.append("&lt;");
		    } else if (c == '&') {
			nval.append("&amp;");
		    } else {			
			nval.append(c);
		    }
		}
		wout.print(nval);
	    } else {
		wout.print(val);
	    }
	    wout.print("</key>");
	}
	wout.print("\n</properties>");
	wout.flush();
    }
    
    /**
     * Searches for the property with the specified key in this property list.
     * If the key is not found in this property list, the default property list,
     * and its defaults, recursively, are then checked. The method returns
     * <code>null</code> if the property is not found.
     *
     * @param   key   the property key.
     * @return  the value in this property list with the specified key value.
     * @see     java.util.Properties#defaults
     * @since   JDK1.0
     */
    public String getProperty(String key) {
	String val = (String)super.get(key);
	return ((val == null) && (defaults != null)) ? defaults.getProperty(key) : val;
    }
    
    /**
     * Searches for the property with the specified key in this property list.
     * If the key is not found in this property list, the default property list,
     * and its defaults, recursively, are then checked. The method returns the
     * default value argument if the property is not found.
     *
     * @param   key            the hashtable key.
     * @param   defaultValue   a default value.
     *
     * @return  the value in this property list with the specified key value.
     * @see     java.util.Properties#defaults
     * @since   JDK1.0
     */
    public String getProperty(String key, String defaultValue) {
	String val = getProperty(key);
	return (val == null) ? defaultValue : val;
    }
    
    /**
     * Returns an enumeration of all the keys in this property list, including
     * the keys in the default property list.
     *
     * @return  an enumeration of all the keys in this property list, including
     *          the keys in the default property list.
     * @see     java.util.Enumeration
     * @see     java.util.Properties#defaults
     * @since   JDK1.0
     */
    public Enumeration propertyNames() {
	Hashtable h = new Hashtable();
	enumerate(h);
	return h.keys();
    }
    
    /**
     * Prints this property list out to the specified output stream. 
     * This method is useful for debugging. 
     *
     * @param   out   an output stream.
     * @since   JDK1.0
     */
    public void list(PrintStream out) {
	out.println("-- listing properties --");
	Hashtable h = new Hashtable();
	enumerate(h);
	for (Enumeration e = h.keys() ; e.hasMoreElements() ;) {
	    String key = (String)e.nextElement();
	    String val = (String)h.get(key);
	    if (val.length() > 40) {
		val = val.substring(0, 37) + "...";
	    }
	    out.println(key + "=" + val);
	}
    }
    
    /**
     * Prints this property list out to the specified output stream. 
     * This method is useful for debugging. 
     *
     * @param   out   an output stream.
     */
    /*
     * Rather than use an anonymous inner class to share common code, this
     * method is duplicated in order to ensure that a non-1.1 compiler can
     * compile this file.
     */
    public void list(PrintWriter out) {
	out.println("-- listing properties --");
	Hashtable h = new Hashtable();
	enumerate(h);
	for (Enumeration e = h.keys() ; e.hasMoreElements() ;) {
	    String key = (String)e.nextElement();
	    String val = (String)h.get(key);
	    if (val.length() > 40) {
		val = val.substring(0, 37) + "...";
	    }
	    out.println(key + "=" + val);
	}
    }
    
    /**
     * Enumerates all key/value pairs in the specified hashtable.
     * @param h the hashtable
     */
    private synchronized void enumerate(Hashtable h) {
	if (defaults != null) {
	    defaults.enumerate(h);
	}
	for (Enumeration e = keys() ; e.hasMoreElements() ;) {
	    String key = (String)e.nextElement();
	    h.put(key, get(key));
	}
    }
}

class PropertiesParser implements DocumentHandler {
    
    // use my own parser if any
    //static final String SAX_DEFAULT = "fr.dyade.koala.xml.sax.SAXParser";
    static final String SAX_DEFAULT = "com.microstar.xml.SAXDriver";
    
    Properties props;
    
    static final int IN_NOTHING  = 0;
    static final int IN_DOCUMENT = 1;
    static final int IN_KEY      = 2;
    int              state       = IN_NOTHING;
    
    String       key;
    StringBuffer value;
    
    Parser       parser;
    
    PropertiesParser(Properties props, InputStream in) 
	throws IOException, SAXException {
	this.props = props;
	
	
        String className = System.getProperty("org.xml.sax.parser");
	
        if (className == null) {
            className = SAX_DEFAULT;
        }
        state = IN_NOTHING;
	value = new StringBuffer();
        try {
            parser = ParserFactory.makeParser(className);
            parser.setDocumentHandler(this);
        } catch (Exception e) {
            e.printStackTrace();
            throw new PropertiesException("can't create parser " + className);
        }
	parser.parse(new InputSource(in));
    }
    
    public void startElement(String name, AttributeList atts)
        throws SAXException {
	if (state == IN_NOTHING) {
	    if (name.equals("properties")) {
		state = IN_DOCUMENT;
	    } else {
		throw new PropertiesException("attempt to find root properties");
	    }
	} else if (state == IN_DOCUMENT) {
	    if (name.equals("key")) {
		state = IN_KEY;
		key = atts.getValue("name");
		
		if (key == null) {
		    throw new PropertiesException("no name for key " + atts);
		}
	    } else {
		throw new PropertiesException("attempt to find keys");
	    }
	} else {
	    throw new PropertiesException("invalid element " + name);
	}
    }
    
    public void endElement(String name) throws SAXException {
	if (state == IN_KEY) {
	    props.put(key, value.toString());
	    state = IN_DOCUMENT;
	    name = null;
	    value = new StringBuffer();
	} else if (state == IN_DOCUMENT) {
	    state = IN_NOTHING;
	}
    }
    
    public void characters(char ch[], int start, int length)
        throws SAXException {
	if (state == IN_KEY) {
	    compute(ch, start, length);
	}
    }
    
    public void ignorableWhitespace(char ch[], int start, int length)
        throws SAXException {
	// nothing to do
    }
    
    public void startDocument() throws SAXException {
	// nothing to do
    }
    
    public void endDocument() throws SAXException {
	// nothing to do
    }
    
    public void processingInstruction(String target, String data)
        throws SAXException {
	// nothing to do
    }
    
    public void setDocumentLocator(Locator locator) {
	// nothing to do
    }
    
    private void compute(char[] ch, int start, int length) {
	int st = start;
	int len = length-1;
	while (st < length 
	       && ((ch[st] == '\n') || (ch[st] == '\t') || (ch[st] == ' ')
		   || (ch[st] == '\r'))) {
	    st++;
	}
	while (len > 0
	       && (ch[len-1] != '\\')
	       && ((ch[len] == '\n') || (ch[len] == '\t') || (ch[len] == ' ')
		   || (ch[len] == '\r'))) {
	    len--;
	}
	
	while (st <= len) {
	    if (ch[st] == '\\') {
		// skip
	    } else {
		value.append(ch[st]);
	    }
	    st++;
	}
    }
}


class PropertiesException extends SAXException {
    PropertiesException(String s) {
	super(s);
    }
}
